# SPDX-FileCopyrightText: 2024-present Sebastian Tobie <sebastian@sebastian-tobie.de>
#
# SPDX-License-Identifier: MIT
import json
import logging
import os.path
import threading
import traceback
from logging.config import dictConfig, fileConfig

from gunicorn.glogging import CONFIG_DEFAULTS
from gunicorn.glogging import Logger as gLogger

__version__ = "0.0.10"


CONFIG_DEFAULTS = {
    "version": 1,
    "disable_existing_loggers": False,
    "root": {"level": "INFO", "handlers": ["console"]},
    "loggers": {},
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "generic",
            "stream": "ext://sys.stdout",
        },
    },
    "formatters": {
        "generic": {
            "format": "%(asctime)s [%(process)d] [%(levelname)s] %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
            "class": "logging.Formatter",
        }
    },
}

REDIRECT_CODES = (301, 302, 303, 307, 308)


class Logger(gLogger):
    access_log_format = "Access(%(s)s) %(r)s"

    def __init__(self, cfg):
        self.logfile = None
        self.lock = threading.Lock()
        self.cfg = cfg
        self.setup(cfg)
        self.error_log = logging.getLogger("gunicorn.error")
        self.access_log = logging.getLogger("gunicorn.access")

    def setup(self, cfg):
        if cfg.logconfig_dict:
            config = CONFIG_DEFAULTS.copy()
            config.update(cfg.logconfig_dict)
            try:
                dictConfig(config)
            except (AttributeError, ImportError, ValueError, TypeError) as exc:
                raise RuntimeError(str(exc))
        elif cfg.logconfig_json:
            config = CONFIG_DEFAULTS.copy()
            if os.path.exists(cfg.logconfig_json):
                try:
                    config_json = json.load(open(cfg.logconfig_json))
                    config.update(config_json)
                    dictConfig(config)
                except (
                    json.JSONDecodeError,
                    AttributeError,
                    ImportError,
                    ValueError,
                    TypeError,
                ) as exc:
                    raise RuntimeError(str(exc))
        elif cfg.logconfig:
            if os.path.exists(cfg.logconfig):
                defaults = CONFIG_DEFAULTS.copy()
                defaults["__file__"] = cfg.logconfig
                defaults["here"] = os.path.dirname(cfg.logconfig)
                fileConfig(
                    cfg.logconfig,
                    defaults=defaults,
                    disable_existing_loggers=False,
                )
            else:
                msg = "Error: log config '%s' not found"
                raise RuntimeError(msg % cfg.logconfig)

    def access(self, resp, req, environ, request_time):
        """See http://httpd.apache.org/docs/2.0/logs.html#combined
        for format details
        """
        level = self.access_log.info
        if resp.status_code >= 400:
            level = self.access_log.warning
        elif resp.status_code >= 500:
            level = self.access_log.error
        safe_atoms = self.atoms_wrapper_class(self.atoms(resp, req, environ, request_time))
        extra = dict(
            METHOD=environ.get("REQUEST_METHOD"),
            PATH=environ.get("PATH_INFO"),
            QUERY=environ.get("QUERY_STRING"),
            PROTOCOL=environ.get("SERVER_PROTOCOL"),
            TIME="%d.%06d" % (request_time.seconds, request_time.microseconds),
            STATUS_CODE=resp.status_code,
            CODE_ARGS=None,
        )
        if environ.get("REMOTE_ADDR", False):
            extra["REMOTE"] = environ.get("REMOTE_ADDR")
        if getattr(resp, "sent", False):
            extra["LENGTH"] = getattr(resp, "sent")
        if environ.get("HTTP_REFERER", False):
            extra["REFERER"] = environ.get("HTTP_REFERER")
        if environ.get("HTTP_USER_AGENT", False):
            extra["USER_AGENT"] = environ.get("HTTP_USER_AGENT")
        if resp.status_code in REDIRECT_CODES:
            for header, value in resp.headers:
                if header == "Location":
                    extra["LOCATION"] = value
                    break
        try:
            level(self.access_log_format, safe_atoms, extra=extra)
        except Exception:
            self.error(traceback.format_exc())
